import { jsx as _jsx } from "react/jsx-runtime";
import { EditorState, StateEffect } from "@codemirror/state";
import { EditorView, tooltips } from "@codemirror/view";
import { memo, useEffect, useMemo, useRef, useState } from "react";
import useInterval from "./useInterval.js";
export const CodeMirror = memo(function CodeMirror(props) {
    const { extensions, style = {}, text, onChange, onFocus, onBlur, autoFocus, putTooltipsInRoot = true } = props;
    const [div, setDiv] = useState();
    const stateRef = useRef();
    const viewRef = useRef();
    const allExtensions = useMemo(() => {
        if (!div) {
            return undefined;
        }
        let allExtensions = extensions.slice();
        allExtensions.push(EditorView.updateListener.of((vu) => {
            if (vu.docChanged) {
                const doc = vu.state.doc;
                const text = doc.toString();
                onChange(text);
            }
            if (vu.focusChanged) {
                if (vu.view.hasFocus) {
                    onFocus && onFocus();
                }
                else {
                    onBlur && onBlur();
                }
            }
        }));
        if (putTooltipsInRoot) {
            // motivation: fixes layout of autocomplete tooltip in notebook-canvas
            const rootNode = div.getRootNode();
            const tooltipParent = rootNode instanceof Document ? document.body : rootNode;
            allExtensions.push(tooltips({ position: 'absolute', parent: tooltipParent }));
        }
        return allExtensions;
    }, [div, extensions, onBlur, onChange, onFocus, putTooltipsInRoot]);
    useEffect(() => {
        if (viewRef.current && allExtensions) {
            viewRef.current.dispatch({ effects: StateEffect.reconfigure.of(allExtensions) });
        }
    }, [allExtensions]);
    useEffect(() => {
        if (div && allExtensions && !stateRef.current) {
            stateRef.current = EditorState.create({
                doc: text,
                extensions: allExtensions,
            });
            viewRef.current = new EditorView({
                state: stateRef.current,
                parent: div,
            });
            if (autoFocus) {
                viewRef.current.focus();
            }
        }
    });
    // TODO: I can't remember exactly why I added this. Looks like it was added when I was working on
    // notebook-canvas. It would be nice to track down exacly why it is needed, and ideally replace it
    // with a more targetted solution.
    useInterval(() => {
        if (viewRef.current) {
            viewRef.current.requestMeasure();
        }
    }, 100);
    // TODO: In the age of reportView, this didn't work. It seems to work now, which is great. But
    // let's monitor it. (Tests?)
    useEffect(() => {
        const view = viewRef.current;
        if (!viewRef.current) {
            return;
        }
        const currentText = view ? view.state.doc.toString() : '';
        if (view && text !== currentText) {
            view.dispatch({
                changes: { from: 0, to: currentText.length, insert: text || '' },
            });
        }
    }, [text]);
    return _jsx("div", { ref: setDiv, className: "cm-theme cm-theme-light", style: style });
});
//# sourceMappingURL=CodeMirror.js.map